/*****************************************************************************
 *  netxplorer.c - SubNetwork Explorer                                       *
 *  Copyright (C) 2000 Prana <pranalukas@gmx.de>                             *
 *  http://cyest.hypermart.net                                               *
 *                                                                           *
 *  This program is free software; you can redistribute it and/or modify     *
 *  it under the terms of the GNU General Public License as published by     *
 *  the Free Software Foundation; either version 2 of the License, or        *
 *  (at your option) any later version.                                      *
 *                                                                           *
 *  This program is distributed in the hope that it will be useful,          *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of           *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            *
 *  GNU General Public License for more details.                             *
 *                                                                           *
 *  You should have received a copy of the GNU General Public License        *
 *  along with this program; if not, write to the Free Software              *
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111 USA      *
 *****************************************************************************/

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <netdb.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netinet/in.h>
//#define DEBUG

int checkftp, checknetbios, checksunrpc;
char *hostname, *progname;
int check_port(const char *ipaddr, int port, char *description) {
/*
   Check existence of port connection
*/
   int sock_descriptor;
   struct sockaddr_in sock;
   socklen_t socklen;
   
   if ((sock_descriptor = socket(PF_INET, SOCK_STREAM, 0)) < 0)
     return(0x10);
   
   memset(&sock,0,sizeof(sock));
   sock.sin_family = AF_INET;
   sock.sin_port = htons(port);
   inet_aton(ipaddr,&sock.sin_addr); // no checking necessary 
   socklen = sizeof(sock);
   if(connect(sock_descriptor, (struct sockaddr *) &sock, socklen))
      return(0x11);
   
   else printf("%s",description);
   return(0);  //Successful
}

int wait_for(int sock,char *expected) {
/*
   Wait and waste garbage
   THIS FUNCTION MIGHT BE BUGGY -- I will fix it later
*/
   char ch[8192];
   int result;
   read(sock,ch,strlen(expected));
#ifdef DEBUG
   	fprintf(stderr,"FTP Code %s\n",ch);
#endif
   
   result =strcmp(expected,ch);
   while (strncmp(ch,"\n",1)) {
        read(sock,ch,1);
   }  
   if (result!=0) result=0;
      else result=1;      
   return result;
}

void send_buffer(int sock, char *buffer) {
/*
   Send buffer ~L~
*/
   write(sock,buffer,strlen(buffer));
}

int ftp_anon(const char *ipaddr) {
/*
   Try to detect anonymous ftp :-)
*/
   int sock_descriptor;
   struct sockaddr_in sock;
   socklen_t socklen;
   int zzz;

   if ((sock_descriptor = socket(PF_INET, SOCK_STREAM, 0)) < 0)
     return(0x10);
   
   memset(&sock,0,sizeof(sock));
   sock.sin_family = AF_INET;
   sock.sin_port = htons(21);
   inet_aton(ipaddr,&sock.sin_addr); // no checking necessary 
   socklen = sizeof(sock);
   if(connect(sock_descriptor, (struct sockaddr *) &sock, socklen))
      return(0x11);  
   else
#ifdef DEBUG
    fprintf(stderr,"%s","FTP service exists\n");
#endif
   if (!wait_for(sock_descriptor,"220"))
      return(0x12); // indicates some error 
   send_buffer(sock_descriptor,"USER anonymous\n");
   if(!wait_for(sock_descriptor,"331"))
      return(0x13);
   send_buffer(sock_descriptor,"PASS britney.spears@my.house\n");
   if(!wait_for(sock_descriptor,"230"))
      return(0x14); // indicate that there is an ftp but non anonymous
   else return(0);
}

int check_host(const char *addr) {
/*
   Check for hostname
*/

  unsigned long qaddr;
  struct hostent *ip_number;
/*  register char **aliases;
  const  char *err_status;*/
  
  qaddr=0;
  qaddr = inet_addr(addr); 
  ip_number = gethostbyaddr((char *)&qaddr,4,AF_INET);
  if (h_errno!=0) {
     h_errno=0;
     return 0;
  }
#ifdef debug
  printf("%s -> %s ",addr, ip_number->h_name);
#else
  printf("%s ",ip_number->h_name); 
#endif
  return (1);
}

void check_ip(int a, int b, int c, int d) {
/*
   Call to hostname
*/
  const char *ip_numbers;
  int ftp_result;
  int z;

  z=0;
  ip_numbers=(char *) malloc(15);
  sprintf((char *)ip_numbers,"%d.%d.%d.%d",a,b,c,d);  
  if (check_host(ip_numbers)) {
    z++;  
    if (checkftp) {
         ftp_result = ftp_anon(ip_numbers);
#ifdef DEBUG
         fprintf(stderr,"FTP Result %d",ftp_result);
#endif
              switch(ftp_result) {
                case 0x00: {printf("anon_ftp ");z++;break;}
                case 0x12: // future 
                case 0x13: // future
                case 0x14: {printf("ftp ");z++;}
              }
    }
  if ((checknetbios) && check_port(ip_numbers,139,"netbios ")) z++;
  if ((checksunrpc) && check_port(ip_numbers,111," sunrpc")) z++;  
  if (z) printf("\n");
  
  }
/*
   Space left for further detection
*/
  free(ip_numbers); 
}

void usage() {
/*
   Display the usage
*/
 char *logo = 
"SubNetwork Explorer v0.01BETA\n\
Copyright (C) 2000 by Prana <pranalukas@gmx.de>\n\
Please visit http://cyest.hypermart.net\n\n";
 
 char *usage= 
 "%s [options] [host range]\n\
 OPTIONS:\n\
 -? : show this help screen\n\
 -f : enable Anonymous ftp detection (port 21) for (ab)use\n\
 -n : enable Netbios file sharing detection (port 139)\n\
 -s : enable SunRPC detection (port 111)\n\
 [host range] : the network you want to scan\n\n\
 EXAMPLE:\n\
 %s 192.168.*.* will display the hostnames from 192.168.0-255.0-255 \n\
 %s -f -n -s 192.168.1.* will scan from 192.168.1.0-255 for Anonymous Ftp, Netbios, SunRPC\n\n\
 NOTE: Host *MUST* be in number format. The result will display the host names though \n";

 printf(logo);
 printf(usage,progname,progname,progname);
 exit(1);
}

void scan_from_third(int a, int b) {
/*
   Heh
*/
 int c, d;

#ifdef DEBUG
  fprintf(stderr, "Scan from third %d.%d.*.*\n",a,b);
#endif
 for (c=0;c<=0xFF;c++)
  if (fork()) {
     for (d=0;d<=0xFF;d++) 
#ifdef DEBUG
      fprintf(stderr,"PID %d ==> %d.%d\n",getpid(),c,d);
#endif
     check_ip(a,b,c,d);
     exit(0);
   } 

}

void scan_from_fourth(int a, int b, int c) {
/*
   it's okay
*/
 int d;

#ifdef DEBUG
 fprintf(stderr,"Scan from fourth %d.%d.%d.*\n",a,b,c);
#endif

   for (d=0;d<=0xFF;d++)
      if (fork()) {
 
#ifdef DEBUG     
        fprintf(stderr,"PID %d ==> %d.%d\n",getpid(),c,d);
#endif
        check_ip(a,b,c,d);
   	exit(0);
   }
 
}

void scan_host(char *host_name) {
/*
   Parse string
*/
   int a,b,c;
   char *s;
   char *cc;
   
   s=(char *)malloc(20);
   s=strcpy(s,host_name);
   cc=strtok(s,".");
   if ((cc==NULL) || (!isdigit(cc[0]))) usage();   
   a=atoi(cc);
   
   cc=strtok(NULL,".");
   if ((cc==NULL) || (!isdigit(cc[0]))) usage();      
   b=atoi(cc);
   
   cc=strtok(NULL,".");
   if (!strcmp(cc,"*")) {
      free(s);   
      scan_from_third(a,b);
      return;
   }
   else c=atoi(cc);
   
   cc=strtok(NULL,".");
   if (!strcmp(cc,"*")) {
      free(s);
      scan_from_fourth(a,b,c);
      }
   else {
      free(s);
      usage();
   }
}

int main(int argc, char *argv[]) {
/*
   Main
*/

  int ch;
  progname=argv[0]; 
 
  while ((ch = getopt(argc, argv, "fnsh:?")) != EOF)
   switch(ch) {
    case 'f':{ checkftp=1;break;}
    case 'n':{ checknetbios=1;break;}
    case 's':{ checksunrpc=1;break;}
    case '?':
    default:usage();
 }

 if (argc <= 1) usage();
   
 argc -= optind;
 argv += optind;
 hostname=*argv;
 scan_host(hostname);
 exit(0);
}
